#Copyright 2009 Anton Vorobyov
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
This script is used to compare two different database versions.
It shows removed/changed/new items with list of changed effects,
changed attributes and effects which were renamed
'''
import sys
sys.path.append("..")

from optparse import OptionParser
from dataFolder import dataFolder
import sqlite3
import os.path
import re

usage = "usage: %prog [--old=OLD] --new=NEW [-ear]"
parser = OptionParser(usage=usage)
parser.add_option("-o", "--old", help="path to old sqlite database (if none specified database currently used by pyfa is taken)")
parser.add_option("-n", "--new", help="path to new sqlite database")
parser.add_option("-e", "--noeffects", action="store_false", dest="effects", help="don't show list of changed effects", default=True)
parser.add_option("-a", "--noattributes", action="store_false", dest="attributes", help="don't show list of changed attributes", default=True)
parser.add_option("-r", "--norenames", action="store_false", dest="renames", help="don't show list of renamed effects", default=True)
(options, args) = parser.parse_args()

#fetch database names from command line arguments
if options.old != None and options.new != None:
    oldDB = sqlite3.connect(options.old)
    newDB = sqlite3.connect(options.new)
elif options.new != None:
    import sql
    sql.connect("sqlite")
    oldDB = sql.getConnection()
    newDB = sqlite3.connect(options.new)
else:
    sys.stderr.write("You need to specify at least 1 database file. Run script with --help option for further info.\n")
    sys.exit()

#initialization of few things used by both changed/renamed effects list
if options.effects or options.renames:
    effectsPath = os.path.join("..", "effects")
    implemented = []

    for filename in os.listdir(effectsPath):
        basename, extension = filename.split('.')
        #Ignore non-py files and exclude 2 pyfa-specific 'effects'
        if extension == "py" and basename != "__init__" and basename != "customEffects":
            implemented.append(basename)

    #effects' names are used w/o &, tab symbols and spaces as pyfa also filters them
    stripSpec = "[&\t ]"
            
    def getEffectStatus(effectName):
        pyfaName = re.sub(stripSpec, "", effectName)
        if pyfaName in implemented: return 'y'
        else: return 'n'

if options.effects or options.attributes:
    #format:
    #key: item id
    #value: [set(effect id 1, effect id 2, ...), {[key: attribute id] | value: (int value, float value)}]
    oldDict = {}
    newDict = {}

    for database, dictionary in ((oldDB, oldDict), (newDB, newDict)):
        if options.effects:
            #Compose list of item IDs (row[0]) and attach list of effect IDs (row[1]) to each
            query = 'SELECT invTypes.typeID, dgmEffects.effectID FROM invTypes INNER JOIN dgmTypeEffects ON dgmTypeEffects.typeID = invTypes.typeID INNER JOIN dgmEffects ON dgmEffects.effectID = dgmTypeEffects.effectID WHERE invTypes.published = 1'
            c = database.cursor()
            c.execute(query)
            for row in c:
                #if there's no such item in dictionary, initialize empty parent list
                if not row[0] in dictionary: dictionary[row[0]] = [set(), {}]
                #add effect to the set
                effectSet = dictionary[row[0]][0]
                effectSet.add(row[1])
        if options.attributes:
            #If there's no list of items, create it and add base and additional attributes to it
            query = 'SELECT dgmTypeAttributes.typeID, dgmTypeAttributes.attributeID, dgmTypeAttributes.valueInt, dgmTypeAttributes.valueFloat, invTypes.mass, invTypes.capacity, invTypes.volume FROM dgmTypeAttributes INNER JOIN invTypes ON dgmTypeAttributes.typeID = invTypes.typeID INNER JOIN invGroups ON invTypes.groupID = invGroups.groupID INNER JOIN invCategories ON invGroups.categoryID = invCategories.categoryID WHERE invTypes.published = 1 AND (invCategories.categoryName = "Ship" OR invCategories.categoryName = "Module" OR invCategories.categoryName = "Charge" OR invCategories.categoryName = "Skill" OR invCategories.categoryName = "Drone" OR invCategories.categoryName = "Implant" OR invCategories.categoryName = "Subsystem" OR invCategories.categoryName = "Celestial")'
            c = database.cursor()
            c.execute(query)
            for row in c:
                #if there's no such item in dictionary, initialize empty parent list
                if not row[0] in dictionary: dictionary[row[0]] = [set(), {}]
                attrDict = dictionary[row[0]][1]
                #add base attributes: mass (4), capacity (38) and volume (161)
                attrDict[4] = (None, row[4])
                attrDict[38] = (None, row[5])
                attrDict[161] = (None, row[6])
                #add non-base attributes to the dictionary
                if not row[1] in (4, 38, 161): attrDict[row[1]] = (row[2], row[3])
                else: print "Warning: base attribute is described in non-base attribute table"

    #Lists contain IDs of items which have the same/changed set of effects' IDs
    same = []
    changed = []

    #iterates through item IDs from old database
    for itemId in oldDict.iterkeys():
        #check if there's such ID in new database
        if itemId in newDict:
            #assume that items are the same, try to prove that it's wrong later
            isSame = True
            oldEffectSet = oldDict[itemId][0]
            newEffectSet = newDict[itemId][0]
            #if item with the same ID and the same set of effects - move to following tests
            if oldEffectSet == newEffectSet:
                oldAttrDict = oldDict[itemId][1]
                newAttrDict = newDict[itemId][1]
                #if set of attributes does not coincide - perform further checks
                if oldAttrDict != newAttrDict:
                    #cycle through attribute list of old items
                    for attributeId in oldAttrDict.iterkeys():
                        #it should be present in new item
                        if attributeId in newAttrDict:
                            #as values are sometimes equal but transferred from int to float and vice versa,
                            #we should take one of them and compare
                            oldTuple = oldAttrDict[attributeId]
                            newTuple = newAttrDict[attributeId]
                            oldVal = oldTuple[0] or oldTuple[1]
                            newVal = newTuple[0] or newTuple[1]
                            if (oldTuple[0] != None and oldTuple[1] != None) or (newTuple[0] != None and newTuple[1] != None):
                                print "Warning: attribute has both int and float values"
                            #eve considers none and zero values as zero
                            for val in (oldVal, newVal):
                                if val == None: val = 0
                            if (oldVal or newVal) and oldVal != newVal:
                                isSame = False
                                break
                        #if there's no such attribute in new item - it's not the same
                        else: isSame = False
                    #if some attribute is present in new item but absent in old, it's not the same
                    for attributeId in newAttrDict.iterkeys():
                        if not attributeId in oldAttrDict:
                            isSame = False
                            break
            #if effects set is not the same - items are not the same
            else: isSame = False
            if isSame: same.append(itemId)
            else: changed.append(itemId)

    #Delete same items from both dictionaries
    for itemId in same:
        del oldDict[itemId], newDict[itemId]

#As pyfa uses effect name as unique ID, we have to keep track of changed effects' names
if options.renames:
    renamed = []
    oldRenDict = {}
    newRenDict = {}

    for database, dictionary in ((oldDB, oldRenDict), (newDB, newRenDict)):
        query = 'SELECT effectID, effectName FROM dgmEffects'
        c = database.cursor()
        c.execute(query)
        for row in c:
            dictionary[row[0]] = re.sub(stripSpec, "", row[1])

    for effectID in oldRenDict.iterkeys():
        if effectID in newRenDict:
            if oldRenDict[effectID] != newRenDict[effectID]:
                renamed.append((oldRenDict[effectID],newRenDict[effectID]))

#Print jobs
if options.effects or options.attributes:
    #print legend only when there're any interesting changes
    if changed:
        print '[+] - new item\n[-] - removed item\n[*] - changed item\n  [+] - effect or attribute has been added to item\n  [-] - effect or attribute has been removed from item\n  [y] - effect is implemented\n  [n] - effect is not implemented\n\nItems:'

        #queries to get item and effect names
        queryTypeName = 'SELECT invTypes.typeName FROM invTypes WHERE invTypes.typeID = ?'
        if options.effects: queryEffectName = 'SELECT dgmEffects.effectName FROM dgmEffects WHERE dgmEffects.effectID = ?'
        if options.attributes: queryAttributeName = 'SELECT dgmAttributeTypes.attributeName FROM dgmAttributeTypes WHERE dgmAttributeTypes.attributeID = ?'

        #process added/removed items; only effect list is printed for ease of maintenance,
        #attributes are not shown as there're usually too many of them
        def printAbsentItems(dict, db, sign):
            for itemId in dict.iterkeys():
                #items which are not changed but remain in old/new list were removed/added
                if not itemId in changed:
                    c = db.cursor()
                    c.execute(queryTypeName, (itemId,))
                    #print item name with added/deleted tag
                    for row in c:
                        print "\n[{0}] {1}".format(sign, row[0])
                    if options.effects:
                        #print effects list
                        for effect in dict[itemId][0]:
                            c = db.cursor()
                            c.execute(queryEffectName, (effect,))
                            for row in c:
                                print "  [{0}|{1}] {2}".format(sign, getEffectStatus(row[0]), re.sub(stripSpec, "",row[0]))

        if options.effects:
            #print old items
            printAbsentItems(oldDict, oldDB, "-")

        #process changed items
        for itemId in changed:
            c = newDB.cursor()
            c.execute(queryTypeName, (itemId,))
            #print item name with changed tag
            for row in c:
                print "\n[*] {0}".format(row[0])

            if options.effects:
                #print effects list
                for dict, compDict, db, sign in ((oldDict, newDict, oldDB, "-"), (newDict, oldDict, newDB, "+")):
                    for effect in dict[itemId][0]:
                        #we don't want to show effects which were unchanged
                        if not effect in compDict[itemId][0]:
                            c = db.cursor()
                            c.execute(queryEffectName, (effect,))
                            for row in c:
                                print "  [{0}|{1}] {2}".format(sign, getEffectStatus(row[0]), re.sub(stripSpec, "",row[0]))

            if options.attributes:
                #prints attributes which are present in baseDict but absent in compDict
                def printAbsentAttrs(baseDict, compDict, db, tag):
                    for attributeId in baseDict.iterkeys():
                        if not attributeId in compDict:
                            tupleValue = baseDict[attributeId][0] or baseDict[attributeId][1]
                            if not tupleValue: tupleValue = 0.0
                            c = db.cursor()
                            c.execute(queryAttributeName, (attributeId,))
                            for row in c:
                                #print zeros instead of none as eve considers none values this way
                                print "  [{0}] {1}: {2}".format(tag, row[0], tupleValue or 0)

                oldAttrDict = oldDict[itemId][1]
                newAttrDict = newDict[itemId][1]

                #seek for removed attributes
                printAbsentAttrs(oldAttrDict, newAttrDict, oldDB, "-")

                #print changed attributes
                for attributeId in oldAttrDict.iterkeys():
                    if attributeId in newAttrDict:
                        if oldAttrDict[attributeId] != newAttrDict[attributeId]:
                            oldVal = oldAttrDict[attributeId][0] or oldAttrDict[attributeId][1]
                            newVal = newAttrDict[attributeId][0] or newAttrDict[attributeId][1]
                            if (oldVal or newVal) and oldVal != newVal:
                                #if attributes exist in both DBs use new one
                                c = newDB.cursor()
                                c.execute(queryAttributeName, (attributeId,))
                                for row in c:
                                    #print zeros instead of none as eve considers none values this way
                                    print "  [*] {0}: {1} => {2}".format(row[0], oldVal or 0, newVal or 0)

                #seek for added attributes
                printAbsentAttrs(newAttrDict, oldAttrDict, newDB, "+")

        if options.effects:
            #print new items
            printAbsentItems(newDict, newDB, "+")

if options.renames:
    if renamed:
        print '\nRenamed effects:'
        for couple in renamed:
            print "\n[{0}] \"{1}\"\n[{2}] \"{3}\"".format(getEffectStatus(couple[0]), couple[0], getEffectStatus(couple[1]), couple[1])
